import logging
import time

from django.core.cache import cache
from django.core.mail import mail_admins
from django.utils.timezone import now


from main.core.exceptions import ParallelExecutionException
from main.core.utils import human_delta

logger = logging.getLogger(__name__)


def avoid_parallel_execution(fn_name):
    def decorator(fn):
        def wrapper(*args, **kwargs):
            key = "double_execution_protection:" + fn_name
            key_value = cache.get(key)

            if key_value is None:
                cache.set(key, now(), None)

                try:
                    fn(*args, **kwargs)
                except BaseException as ex:
                    raise ex
                finally:
                    cache.delete(key)
            else:
                msg = "{0}: parallel execution detected. Other cron started at {1}".format(fn_name, key_value)
                mail_admins("Cron parallel execution detected", msg)
                logger.error(msg)
                raise ParallelExecutionException()

            return None

        return wrapper

    return decorator


def time_perf(desc="Processing", logger=None):
    def decorator(fn):
        def wrapper(*args, **kwargs):
            logger.info(f"{desc}...")
            ts = time.perf_counter()

            result = fn(*args, **kwargs)

            te = time.perf_counter()
            logger.info(f"{desc} done in {human_delta(te - ts)}.")

            return result

        return wrapper

    return decorator
