from rest_framework.filters import OrderingFilter


class NestedOrderingFilter(OrderingFilter):
    """
    Extends default ordering filter to allow sorting by related model attributes and queryset annotations
    without explicitly specifying names for those attributes in ordering_fields
    Desired nesting level can be specified on the view class via order_nesting_depth property
    """

    default_nesting_depth = 2

    def get_nesting_depth(self, view):
        return getattr(view, "order_nesting_depth", self.default_nesting_depth) - 1

    @staticmethod
    def get_field_verbose_name(field, name: str) -> str:
        return field.verbose_name if hasattr(field, "verbose_name") else name.replace("__", " ")

    def add_nested_fields(self, sort_fields, fields, depth: int, parent_name=None):
        for field in fields:
            field_name = f"{parent_name}__{field.name}" if parent_name else field.name
            if field.related_model and (field.many_to_one or field.one_to_one) and depth > 0:
                self.add_nested_fields(sort_fields, field.related_model._meta.get_fields(), depth - 1, field_name)
            elif parent_name:
                sort_fields.append((field_name, self.get_field_verbose_name(field, field_name)))

    def get_valid_fields(self, queryset, view, context=None):
        fields = super(NestedOrderingFilter, self).get_valid_fields(queryset, view, context)
        valid_field_names = {field[0] for field in fields}
        max_depth = self.get_nesting_depth(view)
        model_fields = (field for field in queryset.model._meta.get_fields() if field.name in valid_field_names)
        self.add_nested_fields(fields, model_fields, max_depth)
        # implicitly enable sorting for annotations
        for annotation_name in queryset.query.annotations.keys():
            fields.append((annotation_name, annotation_name))
        return fields
