import abc
import urllib.parse
from contextlib import contextmanager

import requests
from requests.exceptions import ConnectionError, HTTPError, Timeout, TooManyRedirects
from rest_framework.exceptions import ValidationError

from main.core.exceptions import BadGateway, GatewayTimeout
from main.core.utils.classes import Singleton


class ApiClient(Singleton, abc.ABC):
    service_name = "Service"

    def __init__(self):
        self.session = requests.Session()
        self.validate_config()
        self.authorize()

    def get_shared_kwargs(self):
        return {}

    def authorize(self):
        pass

    @abc.abstractmethod
    def get_host(self) -> str:
        pass

    def get_endpoint(self, path) -> str:
        return urllib.parse.urljoin(self.get_host(), path)

    def perform_request(self, method: str, url: str, kwargs):
        with self.handle_errors():
            raise_for_status = kwargs.pop("raise_for_status", True)
            url = self.get_endpoint(url)
            kwargs = {**self.get_shared_kwargs(), **kwargs}
            response = getattr(self.session, method)(url, **kwargs)
            if response.status_code == 401:
                response = self.unauthorized_hook(response)
            if raise_for_status:
                response.raise_for_status()
        return response

    def get(self, url, **kwargs):
        return self.perform_request("get", url, kwargs)

    def post(self, url, **kwargs):
        return self.perform_request("post", url, kwargs)

    def put(self, url, **kwargs):
        return self.perform_request("put", url, kwargs)

    def patch(self, url, **kwargs):
        return self.perform_request("patch", url, kwargs)

    def delete(self, url, **kwargs):
        return self.perform_request("delete", url, kwargs)

    def unauthorized_hook(self, response):
        return response

    @staticmethod
    def validate_config():
        pass

    @classmethod
    @contextmanager
    def handle_errors(cls):
        try:
            yield
        except ConnectionError:
            raise BadGateway("Could not reach %s" % cls.service_name)
        except TooManyRedirects:
            raise BadGateway("Could not reach %s (too many redirects)" % cls.service_name)
        except Timeout:
            raise GatewayTimeout("Request to %s timed out" % cls.service_name)
        except HTTPError as err:
            raise ValidationError(err.response.json())
