import os
import threading

from django.utils import timezone

from main.core.utils.classes import readonly_class_property


class CurrentRequest:
    """
    Contains global state tied to request
    threading.local is not necessary for production, but fixes some `manage.py runsever` inconsistencies
    """

    state = threading.local()

    @classmethod
    def use_request(cls, request):
        cls.state.request = request
        cls.state.cache = {} if request else None

    @readonly_class_property
    def exists(cls):
        return cls.request is not None

    @readonly_class_property
    def request(cls):
        return getattr(cls.state, "request", None)

    @readonly_class_property
    def user(cls):
        if cls.request:
            return cls.request.user
        return None

    @classmethod
    def cached(cls, func):
        def wrapped(*args):
            if getattr(cls.state, "cache", None) is None:
                return func(*args)
            if func not in cls.state.cache:
                cls.state.cache[func] = func(*args)
            return cls.state.cache[func]

        return wrapped


class VLNMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def add_debug_headers(self, response):
        response["X-SME-WORKER-PID"] = os.getpid()
        response["X-SME-WORKER-TIME"] = timezone.now()
        response["X-SME-TIMEZONE"] = timezone.get_default_timezone_name()

    def __call__(self, request):
        CurrentRequest.use_request(request)
        response = self.get_response(request)
        self.add_debug_headers(response)
        CurrentRequest.use_request(None)
        return response
