import uuid

from django.conf import settings
from django.db import models
from main.core.fields import VLNFieldTracker


class TrackerMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        tracker = getattr(self.__class__, "tracker", None)
        if not tracker or not isinstance(tracker, VLNFieldTracker):
            raise AttributeError("Classes with TrackerMixin must have tracker=VLNFieldTracker()")
        self._before_save_hooks = []

    def __flush_before_save_hooks(self):
        for hook in self._before_save_hooks:
            hook()
        self._before_save_hooks.clear()

    def enqueue_before_save_hook(self, hook):
        self._before_save_hooks.append(hook)

    def save(self, *args, **kwargs):
        """
        This implementation was copy pasted from FieldTracker.patch_save
        it correctly reports tracker changes during nested model.save() calls
        """
        self.__flush_before_save_hooks()
        ret = super().save(*args, **kwargs)
        update_fields = kwargs.get("update_fields")
        if not update_fields and update_fields is not None:  # () or []
            fields = update_fields
        elif update_fields is None:
            fields = None
        else:
            fields = (field for field in update_fields if field in self.tracker.fields)
        self.tracker.set_saved_fields(fields=fields)
        return ret

    def save_if_changed(self, **kwargs):
        if not self._state.adding:
            if any(self.tracker.changed().keys()):
                self.save(**kwargs)
        else:
            self.save(**kwargs)


class UUIDIDMixin(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True


class UpdatedAtMixin(models.Model):
    """
    Abstract model which adds update time track.
    """

    updated_at = models.DateTimeField("Updated at", null=True, editable=False, auto_now=True)

    class Meta:
        abstract = True


class CreatedAtMixin(models.Model):
    """
    Abstract model which adds creation time track.
    """

    created_at = models.DateTimeField("Created at", null=True, editable=False, auto_now_add=True)

    class Meta:
        abstract = True


class CreatedByMixin(models.Model):
    """
    Abstract model which adds a track of who created an instance.
    """

    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name="Created by",
        null=True,
        on_delete=models.PROTECT,
        editable=False,
        related_name="%(app_label)s_created_%(class)s_set",
    )

    class Meta:
        abstract = True


class UpdatedByMixin(models.Model):
    """
    Abstract model which adds a track of who updated an instance.
    """

    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name="Updated by",
        null=True,
        on_delete=models.PROTECT,
        editable=False,
        related_name="%(app_label)s_updated_%(class)s_set",
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if isinstance(self, CreatedByMixin) and not self.created_by:
            self.created_by = self.updated_by
        super().save(*args, **kwargs)


class BaseAuditTrack(CreatedAtMixin, UpdatedAtMixin, CreatedByMixin, UpdatedByMixin):
    """
    Abstract model which adds complete audit track.
    """

    class Meta:
        abstract = True


class ArchivableMixin(models.Model):
    """
    Abstract model adding support for 'archive' action (soft delete)
    """

    is_archived = models.BooleanField(verbose_name="Is archived", default=False)

    class Meta:
        abstract = True

    def archive(self, *args, **kwargs):
        self.is_archived = True
        self.save()

    def unarchive(self, *args, **kwargs):
        self.is_archived = False
        self.save()

    def delete(self, *args, **kwargs):
        self.archive(*args, **kwargs)

    def delete_permanent(self, *args, **kwargs):
        super().delete(*args, **kwargs)
