import dataclasses

from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework.fields import get_attribute

from main.core.utils.models import get_model_manager
from main.core.utils.views import extract_id


class ForeignKeyDeserializerMixin:
    queryset = None  # mimics rest_framework.relations.RelatedField

    def __init__(self, *args, **kwargs):
        self.queryset = kwargs.pop("queryset", self.queryset)
        self._get_custom_queryset = kwargs.pop("get_queryset", None)
        super().__init__(*args, **kwargs)

    def get_model_manager(self):
        return get_model_manager(self.Meta.model)

    def _get_queryset(self):
        if self._get_custom_queryset:
            return self._get_custom_queryset(self)
        if self.queryset:
            return self.queryset
        else:
            return self.get_model_manager().all()

    def to_internal_value(self, data):
        pk = extract_id(data)
        try:
            model = self._get_queryset().get(pk=pk)
        except ObjectDoesNotExist:
            raise ValidationError("No matching object found in database")
        return model


class NestedRouterForeignKeyDeserializerMixin:
    """
    This class only works for relations that are serialized as primary keys
    It can be extended if needed or used in conjunction with ForeignKeyDeserializerMixin

    Add your nested router lookups to router_lookups and they will be automatically added to serializer data
    """

    router_lookups = ()

    def _get_relation_pk(self, attr: str):
        view = self.context.get("view")
        return view.kwargs.get(f"{attr}_{view.lookup_field}")

    def _inject_relations_from_router(self, data):
        for attr in self.router_lookups:
            if isinstance(attr, str):
                data[attr] = self._get_relation_pk(attr)
            else:
                data[attr[0]] = self._get_relation_pk(attr[1])
        return data

    def to_internal_value(self, data):
        extended_data = self._inject_relations_from_router(data)
        return super().to_internal_value(extended_data)


class CacheAwareSerializerMixin:
    cache_context_key = "cache"
    cache_pk_attr = "id"
    """
    Can grab model attribute values from context cache by model id instead of querying the model
    Init kwargs:
        consider_empty_on_miss: if true, never try to resolve model attribute on cache miss
    """

    def __init__(self, *args, **kwargs):
        self.consider_empty_on_miss = kwargs.pop("consider_empty_on_miss", False)
        # ListSerializer gets its own set of kwargs, so we need to pluck the attr from child
        if "child" in kwargs:
            self.consider_empty_on_miss = getattr(kwargs["child"], "consider_empty_on_miss", False)
        super().__init__(*args, **kwargs)

    def get_attribute(self, instance):
        try:
            cache = get_attribute(self.context.get(self.cache_context_key), self.source_attrs)
        except (KeyError, AttributeError):
            cache = None
        if cache:
            pk = getattr(instance, self.cache_pk_attr)
            if pk in cache:
                return cache[pk]
            if self.consider_empty_on_miss:
                return [] if self.many else None
        return super().get_attribute(instance)


class GenericDataclassSerializer(serializers.Serializer):  # noqa
    def to_representation(self, instance):
        if not dataclasses.is_dataclass(instance):
            raise ValueError("This serializer only supports dataclasses")
        return dataclasses.asdict(instance)
