import asyncio
import itertools
import numbers
import os
import sys
import time
from collections import defaultdict
from datetime import datetime, timedelta
from functools import cached_property
from inspect import stack
from itertools import groupby
from typing import Callable, Iterable, Iterator

import pytz
from django.conf import settings
from django.db import connection, models, transaction
from django.http import HttpRequest
from django.utils import timezone
from django.utils.duration import _get_duration_components
from rest_framework.request import Request

tz = pytz.timezone(settings.TIME_ZONE)


def get_pk(value):
    if value:
        if isinstance(value, str) and value.isdigit():
            return int(value)
        if isinstance(value, int):
            return value
    return None


def some_none(*args):
    for arg in args:
        if arg is None:
            return True


# Hack
def is_request_from_page_of_model(request, model_name):
    result = False
    http_referer = request.META.get("HTTP_REFERER")
    if http_referer:
        model_name_in_url = "/" + model_name + "/"
        result = model_name_in_url in http_referer
    return result


# Hack
def extract_pk_from_request(request, model_name):
    pk = None
    http_referer = request.META.get("HTTP_REFERER")
    if http_referer:
        splitted_url = http_referer.split("/")[:-1]
        if splitted_url[-3] == model_name:
            pk = get_pk(splitted_url[-2])
    return pk


def get_front_url():
    return _get_url("front")


def get_back_url():
    return _get_url("api")


def get_url_to_logo():
    back = get_back_url()
    return "{0}{1}{2}".format(back, settings.STATIC_URL, "bkb/img/logo.png")


def _get_url(side):
    front_settings = settings.SITES.get(side, {})
    return "{0}://{1}".format(front_settings.get("scheme"), front_settings.get("domain"))


def group_by(data, key_fn):
    """
    Группирует список по ключу
    :param data: список
    :param key_fn: функция вычисления ключа
    :return: list
    """
    grouped_data = defaultdict(list)
    for key, group in groupby(data, key_fn):
        grouped_data[key] += list(group)
    return grouped_data


def get_request_from_stack():
    """
    Tries to find request object among
    frame records of the caller's stack.

    For more information on the interpreter stack, visit
    https://docs.python.org/3.4/library/inspect.html#the-interpreter-stack

    Осторожно! При использовании в рекурсивных вызовах может стать причиной падения производительности

    Returns:
        first found request object or `None`.
    """
    for frame_record in stack():
        request = frame_record[0].f_locals.get("request", None)
        if request and (isinstance(request, Request) or isinstance(request, HttpRequest)):
            return request
    return None


def progress_bar(count, total, status="", bar_len=60):
    """
    Консольный progress bar
    link: https://gist.github.com/vladignatyev/06860ec2040cb497f0f3
    :param count: текущий индекс
    :param total: общее кол-во
    :param status:
    :param bar_len: длина прогесс бара в символах
    :return:
    """
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = "=" * filled_len + "-" * (bar_len - filled_len)

    status = status.strip().replace("\n", " ").replace("\r", "")

    sys.stdout.write("[%s] %s%s ...%s\r" % (bar, percents, "%", status))
    sys.stdout.flush()


def date_to_epoch(date_, set_to_end=False):
    if date_:
        time_ = datetime.max.time() if set_to_end else datetime.min.time()
        datetime_ = datetime.combine(date_, time_)
        date_ = int(datetime_.timestamp())
    return date_


def datetime_to_epoch(datetime_):
    return int(datetime_.timestamp()) if datetime_ else None


def duration_to_epoch(duration):
    result = None
    if isinstance(duration, timedelta):
        result = int(duration.total_seconds())
    return result


def get_datetime_display(datetime_, format_="%Y-%m-%d %H:%M:%S"):
    return timezone.localtime(datetime_).strftime(format_) if datetime_ else None


def get_date_display(date_, format_="%Y-%m-%d"):
    return date_.strftime(format_) if date_ else None


def get_path_to_static_res_by_url(url):
    """Возвращает путь к статическому ресурсу на диске по его url"""
    return os.sep.join((settings.STATIC_ROOT, url.split(settings.STATIC_URL)[1]))


def get_path_to_media_res_by_url(url, filename):
    """Возвращает путь к медиа ресурсу на диске по его url"""
    path = url.split(settings.MEDIA_URL).pop()
    splitted_path = path.split(os.sep)
    splitted_path.pop()
    splitted_path.append(filename)
    return os.sep.join((settings.MEDIA_ROOT, os.sep.join(splitted_path)))


def get_duration_display(duration):
    result = None
    if isinstance(duration, timedelta):
        days, hours, minutes, seconds, microseconds = _get_duration_components(duration)
        result = "{:d}:{:02d}".format(24 * days + hours, minutes)
    return result


def prepare_values_to_serialization(raw_values):
    values = {}
    for key, value in raw_values.items():
        if isinstance(value, timedelta):
            value = get_duration_display(value)
        values[key] = value
    return values


def convert_date_to_datetime(date):
    """Превращает дату в дату время"""
    return datetime.combine(date, datetime.min.time())


def convert_date_to_max_datetime(date):
    """Превращает дату в дату время, возвращает конец дня"""
    return tz.localize(datetime.combine(date, datetime.max.time()))


def human_delta(tdelta):
    """
    Takes a timedelta object and formats it for humans.
    Usage:
        # 149 day(s) 8 hr(s) 36 min 19 sec
        print human_delta(datetime(2014, 3, 30) - datetime.now())
    Example Results:
        23 sec
        12 min 45 sec
        1 hr(s) 11 min 2 sec
        3 day(s) 13 hr(s) 56 min 34 sec
    :param tdelta: The timedelta object or seconds.
    :return: The human formatted timedelta
    """
    if type(tdelta) is float:
        tdelta = timedelta(seconds=tdelta)

    d = dict(days=tdelta.days)
    d["hrs"], rem = divmod(tdelta.seconds, 3600)
    d["min"], d["sec"] = divmod(rem, 60)

    if d["days"] > 0:
        fmt = "{days} day(s) {hrs} hr(s) {min} min {sec} sec"
    elif d["hrs"] > 0:
        fmt = "{hrs} hr(s) {min} min {sec} sec"
    elif d["min"] > 0:
        fmt = "{min} min {sec} sec"
    else:
        fmt = "{sec} sec"

    return fmt.format(**d)


def get_next_id(model):
    db_table = model._meta.db_table
    pk_name = model._meta.pk.column
    sql = """
    SELECT nextval(pg_get_serial_sequence(%s, %s));
    """
    with connection.cursor() as cursor:
        cursor.execute(sql, [db_table, pk_name])
        return cursor.fetchone()[0]


def is_prefetched(obj, related_name):
    """
    Возвращает True, если поле было подгружено через prefetch_related или select_related.
    :param obj: экземпляр модели
    :param related_name: название поля
    :return: bool
    """
    return related_name in getattr(obj, "_prefetched_objects_cache", {}) or related_name in getattr(
        obj._state, "fields_cache", {}
    )  # noqa


def get_fk_field_names(model):
    """Возвращает все названия полей модели, которые являются внешними ключами"""
    return [x.name for x in model._meta.get_fields() if isinstance(x, models.ForeignKey)]


def schedule_task_on_commit(celery_task_fn, *args, **kwargs):
    """
    Асинхронно запускает celery задачу при успешном завершении текущей транзакции
    """
    defer_until_commit(lambda: celery_task_fn.delay(*args, **kwargs))


def defer_until_commit(callback: Callable):
    """
    Вызывает callback при успешном завершении текущей транзакции,
    или сразу, если нет активной транзакции
    """
    if transaction.get_connection().in_atomic_block and not settings.TEST:
        transaction.on_commit(callback)
    else:
        callback()


def get_thread_event_loop():
    try:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    except RuntimeError:
        loop = asyncio.get_event_loop()  # gets previously set event loop, if possible
    return loop


def getattr_nested(obj, path, default=None, separator="."):
    tokens = path.split(separator)
    for token in tokens:
        obj = getattr(obj, token, None)
        if obj is None:
            return default
    return obj


def is_json_primitive(value):
    return (
        value is None
        or isinstance(value, str)
        or isinstance(value, numbers.Number)
        or isinstance(value, bool)
        or isinstance(value, dict)
    )


class writable_cached_property(cached_property):  # noqa
    def __set__(self, instance, value):
        if instance is not None:
            instance.__dict__[self.attrname] = value

    def __delete__(self, instance):
        instance.__dict__.pop(self.attrname, None)


def chunk_iterator(iterable: Iterable, chunk_size: int) -> Iterator[list]:
    iterator = iter(iterable)
    while True:
        chunk = list(itertools.islice(iterator, chunk_size))
        if not chunk:
            return
        yield chunk


class PerfTimer:
    def __init__(self):
        self.total = 0
        self.__start = 0
        self.__last_current = 0

    def __enter__(self):
        self.__start = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__last_current = self.current
        self.total += self.__last_current
        self.__start = 0

    @property
    def current(self):
        if not self.__start:
            return self.__last_current
        return time.perf_counter() - self.__start
