from typing import List

from django.db import models
from django.db.models import prefetch_related_objects

from main.core.utils import is_prefetched, chunk_iterator


def purify_queryset(queryset: models.QuerySet) -> models.QuerySet:
    queryset = queryset.prefetch_related(None).select_related(None)
    queryset.query.annotations.clear()
    return queryset


def get_model_manager(model):
    return model._default_manager or model.objects  # noqa


def patch_attributes(model, **kwargs):
    for key, value in kwargs.items():
        setattr(model, key, value)


def patch_attributes_diff(model, **kwargs) -> bool:
    """
    Patch model attributes and return True if any were changed
    :param model:
    :param kwargs:
    :return: bool
    """
    has_diff = False
    for key, value in kwargs.items():
        has_attr = hasattr(model, key)
        current = getattr(model, key, None)
        if not has_attr or current != value:
            has_diff = True
            setattr(model, key, value)
    return has_diff


def fake_prefetch_related(instance, attribute_name: str, value: List):
    """
    Inserts a list of objects into the models prefetch cache as if they were fetched via prefetch_related()
    :param instance:
    :param attribute_name:
    :param value:
    :return: None
    """
    cache_attr = "_prefetched_objects_cache"
    if not hasattr(instance, cache_attr):
        setattr(instance, cache_attr, {})
    cache = getattr(instance, cache_attr)
    field = getattr(instance, attribute_name)
    prefetch_cache_name = getattr(field, "prefetch_cache_name", cache_attr)
    cache[prefetch_cache_name] = value


def fake_select_related(instance, attribute_name: str, value):
    state = getattr(instance, "_state")
    state.fields_cache[attribute_name] = value


def prefetch_related_iterator(queryset, *related_lookups, chunk_size=100):
    """
    Enables optimization similar to prefetch_related for datasets that may not fit in memory
    :param queryset:
    :param related_lookups:
    :param chunk_size:
    :return:
    """
    qs_iterator = queryset.iterator(chunk_size=chunk_size)
    chunks = chunk_iterator(qs_iterator, chunk_size)
    if not related_lookups:
        related_lookups = getattr(queryset, "_prefetch_related_lookups", [])
    for chunk in chunks:
        prefetch_related_objects(chunk, *related_lookups)
        yield from chunk


class SmartRelatedFilter:
    """
    Фильтр, который можно применять к загружаемым через prefetch_related Queryset'ам без дополнительного запроса к БД.
    На выходе будет массив или queryset в зависимости от того подгружен queryset или нет
    Поддерживаются следующие lookup'ы:
        id=5
        id__in=[5,6,7,8]
    Не поддерживает Q
    Для неподдерживаемых случаев можно использовать custom_object_filter для фильтрации кэшированных queryset'ов
    """

    lookups = {"__in": lambda obj, key, collection: getattr(obj, key[:-4]) in collection}

    @staticmethod
    def default_equals(obj, key, value):
        return getattr(obj, key) == value

    @classmethod
    def lookup_fn_from_key(cls, key: str):
        for lookup_key, lookup_fn in cls.lookups.items():
            if key.endswith(lookup_key):
                return lookup_fn
        return cls.default_equals

    @classmethod
    def apply(cls, model, qs_attribute, *args, custom_object_filter=None, **kwargs):
        if is_prefetched(model, qs_attribute):
            objs = getattr(model, qs_attribute).all()
            if custom_object_filter is not None:
                return [obj for obj in objs if custom_object_filter(obj)]
            return [
                obj
                for obj in objs
                if all(cls.lookup_fn_from_key(key)(obj, key, value) for key, value in kwargs.items())
            ]
        else:
            return getattr(model, qs_attribute).filter(*args, **kwargs)


def smart_exists(model, qs_attribute, **kwargs):
    objs = SmartRelatedFilter.apply(model, qs_attribute, **kwargs)
    if is_prefetched(model, qs_attribute):
        return bool(len(objs))
    else:
        return objs.exists()


def smart_count(model, qs_attribute, **kwargs):
    objs = SmartRelatedFilter.apply(model, qs_attribute, **kwargs)
    if is_prefetched(model, qs_attribute):
        return len(objs)
    else:
        return objs.count()


def BooleanExpression(**kwargs):
    return models.Case(
        models.When(**kwargs, then=models.Value(True)), default=models.Value(False), output_field=models.BooleanField()
    )
