from rest_framework.exceptions import ValidationError
from rest_framework.generics import get_object_or_404

from main.core.utils.models import get_model_manager


def extract_id(data):
    if isinstance(data, int):
        return data
    elif isinstance(data, dict) and "id" in data:
        return data["id"]
    raise ValidationError({"id": "This field is required"})


def find_object_for_data(data, model):
    return get_object_or_404(get_model_manager(model).all(), id=extract_id(data))


def find_objects_for_data(data, model):
    ids = [extract_id(item) for item in data]
    return get_model_manager(model).filter(id__in=ids).all()


def with_headers(response, headers: dict):
    """
    Adds all headers to response
    """
    for key, value in headers.items():
        response[key] = value
    return response


class BodyParser:
    def __init__(self, body: dict):
        self.body = body

    def _get_item(self, name, consume):
        return self.body.pop(name, None) if consume else self.body.get(name, None)

    @classmethod
    def _validate_list(cls, obj, field_name):
        if not isinstance(obj, list):
            raise ValidationError({field_name: "Value must be a list"})

    @classmethod
    def _validate_dict(cls, obj, field_name):
        if not isinstance(obj, dict):
            raise ValidationError({field_name: "Value must be a dict"})

    @classmethod
    def _validate_not_empty(cls, obj, field_name):
        if not obj:
            raise ValidationError({field_name: "No valid items provided"})

    def int_list(self, name: str, not_empty=True, consume=True):
        items = self._get_item(name, consume)
        self._validate_list(items, name)
        items_ints = [item for item in items if isinstance(item, int)]
        if not_empty:
            self._validate_not_empty(items, name)
        return items_ints

    def object_list(self, name: str, not_empty=True, consume=True):
        items = self._get_item(name, consume)
        self._validate_list(items, name)
        if not_empty:
            self._validate_not_empty(items, name)
        return items

    def object_dict(self, name: str, not_empty=True, consume=True):
        object_dict = self._get_item(name, consume)
        self._validate_dict(object_dict, name)
        if not_empty:
            self._validate_not_empty(object_dict, name)
        return object_dict
