from django.db.models import Q
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet, GenericViewSet


class VLNViewMixin:
    @classmethod
    def __normalize_lookup_key(cls, key):
        if not key.endswith("_pk"):
            return key + "_pk"
        return key

    @classmethod
    def __normalize_router_lookup(cls, lookup):
        lookup_pk = lookup
        filter_keys = lookup
        second_filter_keys = lookup
        if not isinstance(lookup, str):
            lookup_pk = lookup[0]
            filter_keys = lookup[1]
            second_filter_keys = lookup[2] if len(lookup) > 2 else None
        lookup_pk = cls.__normalize_lookup_key(lookup_pk)
        return lookup_pk, filter_keys, second_filter_keys

    def get_router_lookup(self, key):
        return self.kwargs.get(self.__normalize_lookup_key(key))

    def get_queryset(self):
        """
        Automatically picks up nested router lookups from both serializer and self
        If serializer has a method to set up prefetching, invokes it automatically with current query set
        :return: queryset
        """
        qs = super().get_queryset()
        serializer_class = self.get_serializer_class()
        if hasattr(serializer_class, "setup_eager_loading"):
            qs = serializer_class.setup_eager_loading(qs)
        router_lookups = getattr(self, "router_lookups", None) or getattr(serializer_class, "router_lookups", None)
        if router_lookups:
            normalized_lookups = (self.__normalize_router_lookup(lookup) for lookup in router_lookups)
            filter_kwargs = {lookup[1]: self.kwargs[lookup[0]] for lookup in normalized_lookups}
            second_filter_kwargs = None
            if len(router_lookups[0]) > 2:
                normalized_lookups = (self.__normalize_router_lookup(lookup) for lookup in router_lookups)
                second_filter_kwargs = {lookup[2]: self.kwargs[lookup[0]] for lookup in normalized_lookups}
            if second_filter_kwargs:
                filter_kwargs = Q(**filter_kwargs)
                filter_kwargs |= Q(**second_filter_kwargs)
                qs = qs.filter(filter_kwargs)
            else:
                qs = qs.filter(**filter_kwargs)
        return qs


class VLNModelViewSet(VLNViewMixin, ModelViewSet):
    pass


class VLNRelatedModelViewSetMixin:
    """
    This mixin relies on VLNViewMixin nested router lookup resolution feature.
    You must specify router lookups for this to work (on serializer or view set).
    Current implementation will only work for one-to-many or many-to-one fields.
    """

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        obj = get_object_or_404(queryset)
        self.check_object_permissions(self.request, obj)
        return obj

    def retrieve_related(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)

    def update_related(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update_related(serializer)
        if getattr(instance, "_prefetched_objects_cache", None):
            instance._prefetched_objects_cache = {}
        return Response(serializer.data)

    def perform_update_related(self, serializer):
        serializer.save()

    def partial_update_related(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.update_related(request, *args, **kwargs)


class VLNRelatedModelViewSet(VLNRelatedModelViewSetMixin, VLNViewMixin, GenericViewSet):
    pass
